#!/usr/bin/env python
# coding=utf-8
# vim: ts=4 sw=4 et
"""
Runs the AHRS calibration online.

Usage:
    ahrs_calibration.py [options]
    ahrs_calibration.py (-h | --help)

Options:
    -l LIGIER_IP    The IP of the ligier [default: 127.0.0.1].
    -p LIGIER_PORT  The port of the ligier [default: 5553].
    -d DET_ID       Detector ID [default: 29].
    -o PLOT_DIR     The directory to save the plot [default: /plots].
    -h --help       Show this screen.

"""
from __future__ import division

from datetime import datetime
from collections import deque, defaultdict
from functools import partial
import io
import os
import random
import time
import threading

import numpy as np
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import matplotlib.dates as md
import seaborn as sns

from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()

import km3db
import km3pipe as kp
from km3pipe.io.daq import TMCHData
from km3modules.common import MemoryObserver
from km3modules.ahrs import fit_ahrs, get_latest_ahrs_calibration
import km3pipe.style

km3pipe.style.use("km3pipe")


class CalibrateAHRS(kp.Module):
    def configure(self):
        self.plots_path = self.require("plots_path")
        det_id = self.require("det_id")
        det_oid = km3db.tools.todetoid(det_id)
        self.time_range = self.get("time_range", default=24 * 3)  # hours
        self.n_points_in_graph = self.get(
            "n_points_in_graph", default=2000
        )  # x-axis resolution (per floor)
        self.data = {}
        self.dus = set()

        self.detector = kp.hardware.Detector(det_id=det_id)
        self.clbmap = km3db.CLBMap(det_oid=det_oid)

        self.cuckoo = kp.time.Cuckoo(60, self.create_plot)  # plot update interval [s]
        self.cuckoo_log = kp.time.Cuckoo(10, self.cprint)
        self.cuckoo_stats = kp.time.Cuckoo(300, self._show_stats)

        # 10Hz monitoring (including bases)
        n_messages = int(
            10 * self.detector.n_doms * self.time_range * 60 ** 2 + self.detector.n_dus
        )

        self.fraction_to_keep = (
            self.n_points_in_graph * self.detector.n_doms / n_messages
        )
        self.cprint(f"Fraction to keep: {self.fraction_to_keep}")
        self.queue_size = int(
            self.n_points_in_graph * 1.1
        )  # a bit of safety margin due to randomness
        self.cprint(
            f"Queue size for each module: {self.queue_size}, time range: {self.time_range} hours"
        )

        self.lock = threading.Lock()
        self.index = 0

    def _show_stats(self):
        """Print some data statistics"""
        messages = ["Recorded data:"]
        for du, data in self.data.items():
            messages.append(f"DU {du}: ")
            for floor, times in data["times"].items():
                messages.append(
                    f"  floor {floor}: {len(times)} ({times[0]}, {times[-1]})"
                )
        self.cprint("\n".join(messages))

    def _register_du(self, du):
        """Create data cache for DU"""
        self.data[du] = {}
        for ahrs_param in ("yaw", "pitch", "roll"):
            self.data[du][ahrs_param] = defaultdict(
                partial(deque, maxlen=self.queue_size)
            )
        self.data[du]["times"] = defaultdict(partial(deque, maxlen=self.queue_size))
        self.dus.add(du)

    def process(self, blob):
        self.index += 1
        if random.random() > self.fraction_to_keep:
            return blob

        self.cuckoo_stats()

        now = datetime.utcnow()
        tmch_data = TMCHData(io.BytesIO(blob["CHData"]))
        dom_id = tmch_data.dom_id
        clb = self.clbmap.dom_ids[dom_id]
        if clb.floor == 0:
            self.log.info("Skipping base CLB")
            return blob

        yaw = tmch_data.yaw
        calib = get_latest_ahrs_calibration(clb.upi, max_version=4)

        if calib is None:
            self.log.warning("No calibration found for CLB UPI '%s'", clb.upi)
            return blob

        du = clb.du
        if du not in self.dus:
            self._register_du(du)
        cyaw, cpitch, croll = fit_ahrs(tmch_data.A, tmch_data.H, *calib)
        self.cuckoo_log(
            "DU{}-DOM{} (random pick): calibrated yaw={}".format(
                clb.du, clb.floor, cyaw
            )
        )
        with self.lock:
            self.data[du]["yaw"][clb.floor].append(cyaw)
            self.data[du]["pitch"][clb.floor].append(cpitch)
            self.data[du]["roll"][clb.floor].append(croll)
            self.data[du]["times"][clb.floor].append(now)

        self.cuckoo.msg()
        return blob

    def create_plot(self):
        self.cprint(self.__class__.__name__ + ": updating plot.")
        if self.time_range > 24:
            xfmt = md.DateFormatter("%Y-%m-%d %H:%M")
        else:
            xfmt = md.DateFormatter("%H:%M")
        xlim = (
            datetime.utcfromtimestamp(time.time() - int(self.time_range * 60 * 60)),
            datetime.utcnow(),
        )
        for du in self.dus:
            data = self.data[du]
            for ahrs_param in data.keys():
                fig, ax = plt.subplots(figsize=(16, 6))
                sns.set_palette("husl", 18)
                ax.set_title(
                    "AHRS {} Calibration on DU{}\n{}".format(
                        ahrs_param, du, datetime.utcnow()
                    )
                )
                ax.set_xlabel("UTC time")
                ax.xaxis.set_major_formatter(xfmt)
                ax.set_ylabel(ahrs_param)
                with self.lock:
                    for floor in sorted(data[ahrs_param].keys()):
                        ax.plot(
                            data["times"][floor],
                            data[ahrs_param][floor],
                            marker=".",
                            linestyle="none",
                            label="Floor {}".format(floor),
                        )
                ax.set_xlim(xlim)
                lgd = plt.legend(bbox_to_anchor=(1.005, 1), loc=2, borderaxespad=0.0)
                fig.tight_layout()
                plt.savefig(
                    os.path.join(
                        self.plots_path, ahrs_param + "_calib_du{}.png".format(du)
                    ),
                    bbox_extra_artists=(lgd,),
                    bbox_inches="tight",
                )
                plt.close("all")


def main():
    from docopt import docopt

    args = docopt(__doc__)

    det_id = int(args["-d"])
    plots_path = args["-o"]
    ligier_ip = args["-l"]
    ligier_port = int(args["-p"])

    pipe = kp.Pipeline()
    pipe.attach(MemoryObserver, every=600000)
    pipe.attach(
        kp.io.ch.CHPump,
        host=ligier_ip,
        port=ligier_port,
        tags="IO_MONIT",
        timeout=60 * 60 * 24 * 7,
        max_queue=2000,
    )
    pipe.attach(CalibrateAHRS, det_id=det_id, plots_path=plots_path)
    pipe.drain()


if __name__ == "__main__":
    main()
